import importlib.resources
import json
from collections import ChainMap

from .broker import BrokerAttributes
from .filter import Filter
from .recommendation import recommendations

IGNORED_CONFIGS = (
    "broker.id",
    "broker.rack",
    "brokerid",
    "logs.dir",
    "zookeeper.connect",
)


def load_defaults(kafka_version: str, source: str):
    return json.load(
        open(
            importlib.resources.files("kafkacfg")
            / f"data/{source}/{kafka_version}.json",
        )
    )


def compute_config_overrides(config: dict, defaults: dict) -> list:
    """Augment each non-default configuration tunable with its associated metadata"""
    overrides = []
    for config_name, config_value in config.items():
        config_default_value = defaults.get(config_name)
        if (
            config_name not in IGNORED_CONFIGS
            and config_default_value is not None
            and config_value != config_default_value["default"]
        ):
            overrides.append(
                {"name": config_name, "override": config_value} | config_default_value
            )
    return overrides


def explain_config(config: dict, defaults: dict) -> list:
    """Augment each configuration tunable with its associated metadata"""
    explained_config = []
    for config_name, config_value in config.items():
        config_default_value = defaults.get(config_name, {})
        explained_config.append(
            {"name": config_name, "override": config_value} | config_default_value
        )
    return explained_config


def filter_config_values(filter_str: str, config: dict, defaults: dict) -> list:
    output = []
    config_filter = Filter.from_str(filter_str)
    for section_defaults in defaults.values():
        matching_configs = {}
        for config_name, config_value in section_defaults.items():
            for predicate in config_filter.predicates:
                full_config_value = config_value | {"name": config_name}
                if not predicate.matches(full_config_value):
                    break
            else:
                matching_configs[config_name] = config.get(config_name)

        if matching_configs:
            output.extend(explain_config(matching_configs, section_defaults))
    return output


def recommend_config(
    config: dict, defaults: dict, broker_attrs: BrokerAttributes
) -> list[str]:
    """Inspect the broker configuration and recommend some configuration tweaks"""
    recos = []
    defaults_kv = {
        cfg_name: cfg_data["default"]
        for cfg_name, cfg_data in defaults["broker"].items()
    }
    broker_config = ChainMap(config, defaults_kv)
    for recommendation in recommendations:
        config_current_value = broker_config.get(recommendation.config)
        config_recommended_value_rendered_formula = recommendation.formula.format(
            **vars(broker_attrs)
        )
        config_recommended_value = eval(config_recommended_value_rendered_formula)
        if not recommendation.operator(
            int(config_current_value), config_recommended_value
        ):
            recos.append(recommendation.render(vars(broker_attrs)))
    return recos


def difference_between_versions(
    from_version: str, to_version: str, config_type: str, source: str
) -> list:
    """Compute the configuration delta between two versions.

    This will highlight the configuration tunables:
    - that have been introduced
    - that have been deprecated
    - which default value has changed

    """
    diff = []
    old_defaults = load_defaults(kafka_version=from_version, source=source)
    new_defaults = load_defaults(kafka_version=to_version, source=source)
    new_flags = set(new_defaults[config_type]) - set(old_defaults[config_type])
    removed_flags = set(old_defaults[config_type]) - set(new_defaults[config_type])
    old_default_col = f"default ({from_version})"
    new_default_col = f"default ({to_version})"

    # Deprecated configs
    for config_name, config_meta in old_defaults[config_type].items():
        if config_name not in removed_flags:
            continue
        config_meta[old_default_col] = config_meta.pop("default")
        config_meta[new_default_col] = "[REMOVED]"
        config_meta = {
            "name": config_name,
            "status": "deprecated",
        } | config_meta  # name 1st means that it will be in the first table column
        diff.append(config_meta)

    # New configs
    for config_name, config_meta in new_defaults[config_type].items():
        if config_name not in new_flags:
            continue
        config_meta[old_default_col] = "[ABSENT]"
        config_meta[new_default_col] = config_meta.pop("default")
        config_meta = {
            "name": config_name,
            "status": "new",
        } | config_meta  # name 1st means that it will be in the first table column
        diff.append(config_meta)

    # Configs with new defaults
    for config_name, config_meta in new_defaults[config_type].items():
        if old_config_meta := old_defaults.get(config_name):
            old_default = old_config_meta["default"]
            new_default = config_meta["default"]
            if old_default != new_default:
                # sometimes old_default = 1234 and new_default = 1234 (xxx)
                if (
                    isinstance(new_default, str)
                    and isinstance(old_default, str)
                    and old_default in new_default
                ):
                    continue
                config_meta[old_default_col] = old_default
                config_meta[new_default_col] = config_meta.pop("default")
                config_meta = (
                    {
                        "name": config_name,
                        "status": "changed",
                    }
                    | config_meta
                )  # name 1st means that it will be in the first table column
                diff.append(config_meta)

    return diff



import re
from dataclasses import dataclass, field

from .exceptions import InvalidPredicateAttribute


@dataclass
class Filter:
    """A filter is simply a list of predicates"""

    predicates: list["Predicate"]

    @classmethod
    def from_str(cls, filter: str) -> "Filter":
        predicates = []
        for predicate_str in filter.split(";"):
            predicates.append(Predicate.from_str(predicate_str))
        return Filter(predicates=predicates)


@dataclass
class Predicate:
    """A predicate equates a field name to a list of values (or patterns)

    Examples:
    - name=log.dirs
    - name=log.dirs,broker.id
    - name=log.*.bytes
    - name=log.*.bytes,socket.*

    """

    attribute: str
    values: list[str] = field(default_factory=list)

    @classmethod
    def from_str(cls, predicate: str) -> "Predicate":
        attribute, values_str = predicate.split("=")
        values = values_str.split(",")
        return Predicate(attribute=attribute, values=values)

    def to_regex(self):
        """Generate a regex in charge of matching any value from self.values"""
        # we convert the human-readable * glob pattern into a regex .*
        values = [val.replace("*", ".*") for val in self.values]
        # we then generate a (x|y|z) regex that can match any of the valuies
        values_pattern = "|".join(values)
        return "(%s)" % (values_pattern)

    def matches(self, config: dict) -> bool:
        if self.attribute not in config:
            raise InvalidPredicateAttribute(
                f"The attribute '{self.attribute}' is not found. "
                f"Available attributes are: {', '.join(config.keys())}"
            )
        attribute_value = config[self.attribute]
        return bool(re.match(self.to_regex(), attribute_value))



import json
from pathlib import Path

import click

from .broker import BrokerAttributes
from .config import (
    compute_config_overrides,
    difference_between_versions,
    explain_config,
    filter_config_values,
    load_defaults,
    recommend_config,
)
from .exceptions import InvalidPredicateAttribute
from .parser import parse_properties_config
from .version import KAFKA_VERSIONS, LIBRDKAFKA_VERSIONS, __version__


def validate_source_and_version(ctx, _, value):
    """Validate the tuple of (source, version) values"""
    source = ctx.params["source"]
    version = value
    if source == "librdkafka" and version not in LIBRDKAFKA_VERSIONS:
        raise click.BadParameter(
            f"{version} is not a valid librdkafka version. Choose from {', '.join(LIBRDKAFKA_VERSIONS)}"
        )
    elif source == "kafka" and version not in KAFKA_VERSIONS:
        raise click.BadParameter(
            f"{version} is not a valid kafka version. Choose from {', '.join(KAFKA_VERSIONS)}"
        )
    return value


def validate_config_type_and_source(ctx, _, value):
    """Validate the tuple of (source, config_type) values"""
    source = ctx.params["source"]
    config_type = value
    if source == "librdkafka" and config_type in ("broker", "topic"):
        raise click.BadParameter(
            f"Invalid librdkafka config type '{config_type}'. Choose from 'consumer' or 'producer'"
        )
    return value


def config_type_default():
    ctx = click.get_current_context()
    source = ctx.params["source"]
    if source == "librdkafka":
        return "producer"
    else:
        return "broker"


version_choice = click.option(
    "-v",
    "--version",
    help="The version of the software associated to the configuration source (kafka, librdkafka) you are running",
    callback=validate_source_and_version,
    required=True,
)

config_type_choice = click.option(
    "-c",
    "--config-type",
    type=click.Choice(("broker", "consumer", "producer", "topic")),
    help="The type of configuration to evaluate",
    callback=validate_config_type_and_source,
    default=lambda: config_type_default(),
    show_default=True,
)

source_choice = click.option(
    "-s",
    "--source",
    type=click.Choice(("kafka", "librdkafka")),
    help="The source of configuration to evaluate: the kafka or librdkafka documentation",
    default="kafka",
    # useful to always have ctx.params['source'] in validate_source_version,
    # whatever the order with which --source and --version were passed in
    is_eager=True,
    show_default=True,
)


@click.group()
@click.version_option(__version__)
def kafkacfg():
    """A kafka configuration inspector"""


@kafkacfg.command()
@source_choice
@version_choice
@config_type_choice
@click.argument("config_file", type=click.Path(exists=True))
def overrides(
    source: str,
    version: str,
    config_file: click.Path,
    config_type: str,
):
    """Display the config overrides from a kafka configuration file"""
    config = parse_properties_config(Path(config_file))
    defaults = load_defaults(version, source=source)[config_type]
    config_overrides = compute_config_overrides(config, defaults)
    click.echo(json.dumps(config_overrides))


@kafkacfg.command()
@source_choice
@version_choice
@config_type_choice
@click.argument("config_file", type=click.Path(exists=True))
def explain(
    source: str,
    version: str,
    config_file: click.Path,
    config_type: str,
):
    """Display information about each config tunable from a kafka configuration file"""
    config = parse_properties_config(Path(config_file))
    defaults = load_defaults(version, source=source)[config_type]
    config_overrides = explain_config(config, defaults)
    click.echo(json.dumps(config_overrides))


@kafkacfg.command()
@source_choice
@version_choice
@click.option("-c", "--config-file", type=click.Path(exists=True))
@click.argument("query", type=str)
def query(source: str, version: str, query: str, config_file: click.Path):
    """Query kafka configuration tunables by their attributes

    Examples:

    \b
    # Query a kafka config details by tunable name
    $ kafkacfg query -v 1.1 'name=replica.fetch.min.bytes'
    \b
    # Query multiple kafka config details by tunable name
    $ kafkacfg query -v 1.1 'name=replica.fetch.min.bytes,socket.request.max.bytes'

    \b
    # Query multiple kafka config details by tunable name pattern
    $ kafkacfg query -v 1.1 'name=socket.*.buffer.bytes'

    \b
    # Query a kafka config details by importance
    $ kafkacfg query -v 1.1 'importance=high'

    \b
    # Query multiple kafka config details by importance and name pattern
    $ kafkacfg query -v 1.1 'importance=high;name=*.bytes'

    \b
    # Query kafka confifg details and compare results with config overrides
    $ kafkacfg query -v 1.1 -c server.properties 'importance=high;name=*.bytes'

    \b
    # Query librdkafka confifg details and compare results with config overrides
    $ kafkacfg query --source librdkafka -v 2.0.0  'importance=high;name=*.bytes'

    """
    config = parse_properties_config(Path(config_file)) if config_file else {}
    defaults = load_defaults(version, source=source)
    try:
        filtered_configs = filter_config_values(query, config, defaults)
    except InvalidPredicateAttribute as exc:
        click.echo(exc, err=True)
    else:
        click.echo(json.dumps(filtered_configs))


@kafkacfg.command()
@version_choice
@click.option(
    "--broker-num-cpus",
    type=int,
    help="The number of processors per broker",
    required=True,
)
@click.option(
    "--broker-num-disks",
    type=int,
    help="The number of hard drives per broker",
    required=True,
)
@click.argument("config_file", type=click.Path(exists=True))
def recommends(
    version: str,
    config_file: click.Path,
    broker_num_cpus: int,
    broker_num_disks: int,
):
    """Emit sourced configuration recommendations based on broker attributes"""
    config = parse_properties_config(Path(config_file))
    defaults = load_defaults(version, source="kafka")
    broker_attrs = BrokerAttributes(
        num_cpus=broker_num_cpus, num_disks=broker_num_disks
    )
    recos = recommend_config(config, defaults, broker_attrs)
    for i, reco in enumerate(recos, 1):
        click.echo(f"Recommendation {i}: set {reco}")


@kafkacfg.command()
@source_choice
@click.option(
    "-f",
    "--from-version",
    callback=validate_source_and_version,
    help="The initial version",
)
@click.option(
    "-t",
    "--to-version",
    callback=validate_source_and_version,
    help="The destination version",
)
@config_type_choice
def diff(source: str, from_version: str, to_version: str, config_type: str):
    """Display the changes in configuration between 2 kafka versions"""
    click.echo(
        json.dumps(
            difference_between_versions(
                from_version, to_version, config_type, source=source
            )
        )
    )



import operator as op
from dataclasses import dataclass, field

operator_repr = {
    op.eq: "==",
    op.ne: "!=",
    op.ge: ">=",
    op.gt: ">",
    op.le: "<=",
    op.lt: "<",
}


@dataclass
class Recommendation:
    config: str
    formula: str
    operator: callable
    sources: list[str] = field(default_factory=list)

    def __str__(self):
        return f"{self.config} {operator_repr[self.operator]} {self.formula}"

    def evaluate(self, context: dict) -> str:
        rendered_formula = self.formula.format(**context)
        evaluated_formula = eval(rendered_formula)
        parts = [self.config, operator_repr[self.operator], evaluated_formula]
        if str(evaluated_formula) != str(self.formula):
            parts.append(f"({self.formula.replace('{', '').replace('}', '')})")
        return " ".join(map(str, parts))

    def render(self, context: dict) -> str:
        evaluation = self.evaluate(context)
        sources = "\n".join([f"* {source}" for source in self.sources])
        return f"{evaluation}\n{sources}"


recommendations = [
    Recommendation(
        config="num.io.threads",
        formula="{num_disks}",
        operator=op.eq,
        sources=["https://strimzi.io/blog/2021/06/08/broker-tuning/"],
    ),
    Recommendation(
        config="replica.fetch.min.bytes",
        formula="512",
        operator=op.eq,
        sources=[
            "https://azure.microsoft.com/en-us/blog/processing-trillions-of-events-per-day-with-apache-kafka-on-azure/"
        ],
    ),
    Recommendation(
        config="socket.receive.buffer.bytes",
        formula="512",
        operator=op.eq,
        sources=[
            "https://azure.microsoft.com/en-us/blog/processing-trillions-of-events-per-day-with-apache-kafka-on-azure/",
            "https://strimzi.io/blog/2021/06/08/broker-tuning/",
        ],
    ),
    Recommendation(
        config="socket.send.buffer.bytes",
        formula="1048576",
        operator=op.eq,
        sources=[
            "https://azure.microsoft.com/en-us/blog/processing-trillions-of-events-per-day-with-apache-kafka-on-azure/",
            "https://strimzi.io/blog/2021/06/08/broker-tuning/",
        ],
    ),
    Recommendation(
        config="replica.socket.receive.buffer.bytes",
        formula="1048576",
        operator=op.eq,
        sources=[
            "https://azure.microsoft.com/en-us/blog/processing-trillions-of-events-per-day-with-apache-kafka-on-azure/"
        ],
    ),
    Recommendation(
        config="socket.request.max.bytes",
        formula="104857600",
        operator=op.eq,
        sources=[
            "https://azure.microsoft.com/en-us/blog/processing-trillions-of-events-per-day-with-apache-kafka-on-azure/"
        ],
    ),
]



import configparser
from pathlib import Path


def parse_properties_config(filepath: Path) -> dict:
    """Parse a .properties file and return its content as a dict

    Note: a properties file is just an ini file without a global section heading.

    """
    config_text = f"[{configparser.DEFAULTSECT}]\n" + filepath.read_text()
    config = configparser.ConfigParser()
    config.read_string(config_text)
    return dict(config[configparser.DEFAULTSECT])
